class AerieException(Exception):
    pass


class DriverNotRegistered(AerieException):
    pass


class IntegrityError(AerieException):
    pass


class UniqueViolationError(IntegrityError):
    pass



import typing as t
import urllib.parse


class URL:
    def __init__(self, url: str):
        self.url = url
        self._components: urllib.parse.ParseResult = urllib.parse.urlparse(url)

    @property
    def scheme(self) -> str:
        return self._components.scheme

    @property
    def driver(self) -> str:
        return self.scheme.split("+")[0]

    @property
    def library(self) -> t.Optional[str]:
        if "+" in self.scheme:
            return self.scheme.split("+")[1]
        return None

    @property
    def hostname(self) -> t.Optional[str]:
        return self._components.hostname

    @property
    def port(self) -> t.Optional[int]:
        return self._components.port

    @property
    def username(self) -> t.Optional[str]:
        return self._components.username

    @property
    def password(self) -> t.Optional[str]:
        return self._components.password

    @property
    def db_name(self) -> str:
        if self.driver == "sqlite":
            return self._components.path
        return self._components.path[1:]

    @property
    def options(self) -> dict:
        return dict(urllib.parse.parse_qsl(self._components.query))

    def __str__(self):
        credentials = ""
        if self.username:
            credentials = f"{self.username}:******@"

        components = self._components._replace(
            netloc=f"{credentials}{self.hostname}:{self.port}"
        )
        return urllib.parse.urlunparse(components)



from __future__ import annotations

import sqlite3
import typing as t
import uuid

import aiosqlite
import pypika as pk

from aerie.exceptions import UniqueViolationError
from aerie.protocols import BaseConnection, BaseDriver, BaseSavePoint, BaseTransaction
from aerie.terms import OnConflict
from aerie.url import URL


class _Pool:
    def __init__(self, url: URL, **options: t.Any) -> None:
        self._url = url
        self._options = options

    async def acquire(self) -> aiosqlite.Connection:
        connection = aiosqlite.connect(
            database=self._url.db_name, isolation_level=None, **self._options
        )
        await connection.__aenter__()
        connection.row_factory = aiosqlite.Row
        return connection

    async def release(self, connection: aiosqlite.Connection) -> None:
        await connection.__aexit__(None, None, None)


class _SavePoint(BaseSavePoint):
    def __init__(self, connection: _Connection, name: str = None) -> None:
        self.connection = connection
        self.name = name or "aerie_" + str(uuid.uuid4()).replace("-", "")

    async def begin(self) -> _SavePoint:
        await self.connection.execute(f"SAVEPOINT {self.name}")
        return self

    async def commit(self) -> None:
        await self.connection.execute(f"RELEASE SAVEPOINT {self.name}")

    async def rollback(self) -> None:
        await self.connection.execute(f"ROLLBACK TO SAVEPOINT {self.name}")


class _Transaction(BaseTransaction):
    def __init__(self, connection: _Connection) -> None:
        self.connection = connection
        self._savepoint = _SavePoint(self.connection)
        self._is_root = True

    async def begin(self, is_root: bool = True) -> _Transaction:
        self._is_root = is_root
        if self._is_root:
            await self.connection.execute("BEGIN")
        else:
            await self._savepoint.begin()
        return self

    async def commit(self) -> None:
        if self._is_root:
            await self.connection.execute("COMMIT")
        else:
            await self._savepoint.commit()

    async def rollback(self) -> None:
        if self._is_root:
            await self.connection.execute("ROLLBACK")
        else:
            await self._savepoint.rollback()


class _Connection(BaseConnection):
    def __init__(self, pool: _Pool) -> None:
        self._pool = pool
        self._connection: t.Optional[aiosqlite.Connection] = None

    async def acquire(self) -> None:
        self._connection = await self._pool.acquire()

    async def release(self) -> None:
        await self._pool.release(self._connection)

    async def execute(self, stmt: str, params: t.Mapping = None) -> t.Any:
        assert self._connection is not None, "Connection is not acquired."
        try:
            cursor = await self._connection.execute(stmt, params)
            if cursor.lastrowid == 0:
                return cursor.rowcount
            return cursor.lastrowid
        except sqlite3.IntegrityError as ex:
            raise UniqueViolationError(str(ex)) from ex

    async def execute_all(
        self,
        stmt: str,
        params: t.List[t.Mapping] = None,
    ) -> t.Any:
        assert self._connection is not None, "Connection is not acquired."
        async with self._connection.executemany(stmt, params) as cursor:
            if cursor.lastrowid == 0:
                return cursor.rowcount
            return cursor.lastrowid

    async def fetch_one(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.Optional[t.Mapping]:
        assert self._connection is not None, "Connection is not acquired."
        async with self._connection.execute(stmt, params) as cursor:
            return await cursor.fetchone()

    async def fetch_all(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.List[t.Mapping]:
        assert self._connection is not None, "Connection is not acquired."
        async with self._connection.execute(stmt, params) as cursor:
            return await cursor.fetchall()

    async def iterate(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.AsyncGenerator[t.Any, None]:
        assert self._connection is not None, "Connection is not acquired."
        async with self._connection.execute(stmt, params) as cursor:
            async for row in cursor:
                yield row

    def transaction(self) -> _Transaction:
        return _Transaction(self)

    @property
    def raw_connection(self) -> aiosqlite.Connection:
        return self._connection

    async def __aenter__(self) -> _Connection:
        await self.acquire()
        return self

    async def __aexit__(self, *args) -> None:
        await self.release()


class SQLiteQuery(pk.queries.Query):
    @classmethod
    def _builder(cls, **kwargs) -> SQLiteQueryBuilder:
        return SQLiteQueryBuilder(**kwargs)


class SQLiteQueryBuilder(pk.queries.QueryBuilder):
    def __init__(self, **kwargs: str) -> None:
        super().__init__(dialect=pk.dialects.Dialects.SQLLITE, **kwargs)
        self._on_conflict = False
        self._on_conflict_fields: t.List[pk.queries.Field] = []
        self._on_conflict_do_nothing = False
        self._on_conflict_do_updates: t.List[t.Tuple] = []
        self._on_conflict_wheres = None
        self._on_conflict_do_update_wheres = None

    @pk.utils.builder
    def on_conflict(
        self,
        *target_fields: t.Union[str, pk.terms.Term],
    ) -> SQLiteQueryBuilder:
        if not self._insert_table:
            raise pk.queries.QueryException("On conflict only applies to insert query")

        self._on_conflict = True

        for target_field in target_fields:
            if isinstance(target_field, str):
                self._on_conflict_fields.append(self._conflict_field_str(target_field))
            elif isinstance(target_field, pk.terms.Term):
                self._on_conflict_fields.append(target_field)
        return self

    @pk.utils.builder
    def do_nothing(self) -> SQLiteQueryBuilder:
        if len(self._on_conflict_do_updates) > 0:
            raise pk.queries.QueryException("Can not have two conflict handlers")
        self._on_conflict_do_nothing = True
        return self

    @pk.utils.builder
    def do_update(
        self,
        update_field: t.Union[str, pk.queries.Field],
        update_value: t.Optional[t.Any] = None,
    ) -> SQLiteQueryBuilder:
        if self._on_conflict_do_nothing:
            raise pk.queries.QueryException("Can not have two conflict handlers")

        if isinstance(update_field, str):
            field = self._conflict_field_str(update_field)
        elif isinstance(update_field, pk.terms.Field):
            field = update_field
        else:
            raise pk.queries.QueryException("Unsupported update_field")

        if update_value is not None:
            self._on_conflict_do_updates.append(
                (field, pk.terms.ValueWrapper(update_value))
            )
        else:
            self._on_conflict_do_updates.append((field, None))
        return self

    def _conflict_field_str(self, term: str) -> t.Optional[pk.terms.Field]:
        if self._insert_table:
            return pk.terms.Field(term, table=self._insert_table)
        return None

    def _on_conflict_sql(self, **kwargs: t.Any) -> str:
        if not self._on_conflict_do_nothing and len(self._on_conflict_do_updates) == 0:
            if not self._on_conflict_fields:
                return ""
            raise pk.queries.QueryException("No handler defined for on conflict")

        if self._on_conflict_do_updates and not self._on_conflict_fields:
            raise pk.queries.QueryException(
                "Can not have fieldless on conflict do update"
            )

        conflict_query = " ON CONFLICT"
        if self._on_conflict_fields:
            fields = [
                f.get_sql(with_alias=True, **kwargs) for f in self._on_conflict_fields
            ]
            conflict_query += " (" + ", ".join(fields) + ")"

        if self._on_conflict_wheres:
            conflict_query += " WHERE {where}".format(
                where=self._on_conflict_wheres.get_sql(subquery=True, **kwargs)
            )

        return conflict_query

    def _on_conflict_action_sql(self, **kwargs: t.Any) -> str:
        if self._on_conflict_do_nothing:
            return " DO NOTHING"
        elif len(self._on_conflict_do_updates) > 0:
            updates = []
            for field, value in self._on_conflict_do_updates:
                if value:
                    updates.append(
                        "{field}={value}".format(
                            field=field.get_sql(**kwargs),
                            value=value.get_sql(with_namespace=True, **kwargs),
                        )
                    )
                else:
                    updates.append(
                        "{field}=EXCLUDED.{value}".format(
                            field=field.get_sql(**kwargs),
                            value=field.get_sql(**kwargs),
                        )
                    )
            action_sql = " DO UPDATE SET {updates}".format(updates=",".join(updates))

            if self._on_conflict_do_update_wheres:
                action_sql += " WHERE {where}".format(
                    where=self._on_conflict_do_update_wheres.get_sql(
                        subquery=True, with_namespace=True, **kwargs
                    )
                )
            return action_sql

        return ""

    def get_sql(
        self, with_alias: bool = False, subquery: bool = False, **kwargs: t.Any
    ) -> str:
        self._set_kwargs_defaults(kwargs)

        querystring = super().get_sql(with_alias, subquery, **kwargs)

        querystring += self._on_conflict_sql(**kwargs)
        querystring += self._on_conflict_action_sql(**kwargs)

        return querystring


class SQLiteDriver(BaseDriver[SQLiteQuery, SQLiteQueryBuilder]):
    dialect = "sqlite"
    can_create_database = False
    query_class = SQLiteQuery

    def __init__(self, url: URL) -> None:
        self.url = url
        self.pool = _Pool(url)

    async def connect(self) -> None:
        pass

    async def disconnect(self) -> None:
        pass

    def connection(self) -> _Connection:
        return _Connection(self.pool)

    def insert_query(
        self,
        table_name: str,
        values: t.Mapping,
        on_conflict: t.Union[str, t.List[str]] = OnConflict.RAISE,
        conflict_target: t.Union[str, t.List[str]] = None,
        replace_except: t.List[str] = None,
    ) -> SQLiteQueryBuilder:
        qb = super().insert_query(table_name, values)
        conflict_target = conflict_target or []
        replace_except = replace_except or []

        if on_conflict != OnConflict.RAISE:
            qb = qb.on_conflict(*conflict_target)

        if on_conflict == OnConflict.NOTHING:
            qb = qb.do_nothing()

        if on_conflict == OnConflict.REPLACE:
            for column, value in values.items():
                if column in replace_except:
                    continue
                qb = qb.do_update(column, value)
        return qb



from __future__ import annotations

import typing as t

import pypika as pk
import pypika.functions as fn

from aerie.terms import OnConflict, Raw


class StrLike(t.Protocol):
    def __str__(self) -> str:
        ...


Queryable = t.Union[str, StrLike]


class Row(t.Mapping):
    pass


class BaseConnection:
    async def acquire(self) -> None:
        raise NotImplementedError()

    async def release(self) -> None:
        raise NotImplementedError()

    async def execute(self, stmt: str, params: t.Mapping = None) -> t.Any:
        raise NotImplementedError()

    async def execute_all(
        self,
        stmt: str,
        params: t.List[t.Mapping] = None,
    ) -> t.Any:
        raise NotImplementedError()

    async def fetch_one(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.Optional[t.Mapping]:
        raise NotImplementedError()

    async def fetch_val(
        self, stmt: str, params: t.Mapping = None, column: t.Any = 0
    ) -> t.Optional[t.Any]:
        row = await self.fetch_one(stmt, params)
        return None if row is None else row[column]

    async def fetch_all(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.List[t.Mapping]:
        raise NotImplementedError()

    async def iterate(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.AsyncGenerator[t.Mapping, None]:
        raise NotImplementedError()
        # noinspection PyUnreachableCode
        yield True  # pragma: nocover

    def transaction(self) -> BaseTransaction:
        raise NotImplementedError()

    @property
    def raw_connection(self) -> t.Any:
        raise NotImplementedError()


class BaseSavePoint:
    async def begin(self) -> BaseSavePoint:
        raise NotImplementedError()

    async def commit(self):
        raise NotImplementedError()

    async def rollback(self):
        raise NotImplementedError()


class BaseTransaction:
    async def begin(self, is_root: bool = True) -> BaseTransaction:
        raise NotImplementedError()

    async def commit(self):
        raise NotImplementedError()

    async def rollback(self):
        raise NotImplementedError()


Q = t.TypeVar("Q", bound=pk.queries.Query)
QB = t.TypeVar("QB", bound=pk.queries.QueryBuilder)

IterableValues = t.Union[
    t.List[t.Mapping],
    t.Tuple[t.Mapping],
]


class BaseDriver(t.Generic[Q, QB]):
    dialect: str = "unknown"
    can_create_database: bool = True
    query_class: t.Type[QB] = pk.queries.Query

    async def connect(self):
        raise NotImplementedError()

    async def disconnect(self):
        raise NotImplementedError()

    def connection(self) -> BaseConnection:
        raise NotImplementedError()

    def get_query_builder(self) -> Q:
        return self.query_class()

    def count_query(
        self,
        table_name: str,
        column: str = "*",
        where: t.Union[str, pk.terms.Term] = None,
    ) -> QB:
        query = self.get_query_builder()
        query = query.from_(table_name).select(fn.Count(column))
        if where:
            if isinstance(where, str):
                query = query.where(Raw(where))
            else:
                query = query.where(where)
        return query

    def insert_query(
        self,
        table_name: str,
        values: t.Mapping,
        on_conflict: str = OnConflict.RAISE,
        conflict_target: t.Union[str, t.List[str]] = None,
        replace_except: t.List[str] = None,
    ) -> QB:
        qb = self.get_query_builder()
        columns = values.keys()
        _values = [pk.NamedParameter(column) for column in columns]
        return qb.into(table_name).columns(*columns).insert(*_values)

    def insert_all_query(
        self,
        table_name: str,
        values: IterableValues,
    ) -> QB:
        values_0 = values[0]
        return self.insert_query(table_name, values_0)

    def update_query(
        self,
        table_name: str,
        values: t.Mapping,
        where: t.Union[str, pk.terms.Term] = None,
    ) -> QB:
        qb = self.get_query_builder()
        update = qb.update(table_name)
        for key, value in values.items():
            update = update.set(key, pk.NamedParameter(key))

        if where:
            if isinstance(where, str):
                update = update.where(Raw(where))
            else:
                update = update.where(where)
        return update

    def delete_query(
        self,
        table_name: str,
        where: t.Union[str, pk.terms.Term] = None,
    ) -> QB:
        qb = self.get_query_builder()
        query = qb.from_(table_name).delete()
        if where:
            if isinstance(where, str):
                query = query.where(Raw(where))
            else:
                query = query.where(where)
        return query



import importlib
import typing as t


def import_string(path: str) -> t.Any:
    module_name, *rest, member = path.rpartition(".")
    module = importlib.import_module(module_name)
    return getattr(module, member)



from __future__ import annotations

import itertools
import typing as t
from types import TracebackType

import asyncpg
import asyncpg.pool
import pypika as pk

from aerie.exceptions import UniqueViolationError
from aerie.protocols import BaseConnection, BaseDriver, BaseTransaction
from aerie.terms import OnConflict
from aerie.url import URL


class _Transaction(BaseTransaction):
    def __init__(self, connection: _Connection) -> None:
        self.connection = connection

    async def begin(self, is_root: bool = True) -> _Transaction:
        self._tx = self.connection.raw_connection.transaction()
        await self._tx.start()
        return self

    async def commit(self) -> None:
        await self._tx.commit()

    async def rollback(self) -> None:
        await self._tx.rollback()

    async def __aenter__(self):
        return await self.begin()

    async def __aexit__(
        self,
        exc_type: t.Tuple[BaseException],
        exc_val: BaseDriver,
        exc_tb: TracebackType,
    ):
        if exc_type is not None:
            await self.rollback()
        else:
            await self.commit()


class _Connection(BaseConnection):
    def __init__(self, pool: asyncpg.pool.Pool) -> None:
        self._pool = pool
        self._connection: t.Optional[asyncpg.connection.Connection] = None

    async def acquire(self) -> None:
        await self._pool
        self._connection = await self._pool.acquire()

    async def release(self) -> None:
        await self._pool.release(self._connection)

    async def execute(self, stmt: str, params: t.Mapping = None) -> t.Any:
        assert self._connection is not None, "Connection is not acquired."
        stmt, args = self._replace_placeholders(stmt, params)
        try:
            return await self._connection.fetchval(stmt, *args)
        except asyncpg.UniqueViolationError as ex:
            raise UniqueViolationError(str(ex)) from None

    async def execute_all(
        self,
        stmt: str,
        params: t.List[t.Mapping] = None,
    ) -> t.Any:
        assert self._connection is not None, "Connection is not acquired."
        _args = []
        if params:
            for index, data in enumerate(params):
                if index == 0:
                    stmt, args_0 = self._replace_placeholders(stmt, data)
                    _args.append(args_0)
                else:
                    _args.append(list(data.values()))

        try:
            return await self._connection.executemany(stmt, _args)
        except asyncpg.UniqueViolationError as ex:
            raise UniqueViolationError(str(ex)) from None

    async def fetch_one(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.Optional[t.Mapping]:
        assert self._connection is not None, "Connection is not acquired."
        stmt, args = self._replace_placeholders(stmt, params)
        return await self._connection.fetchrow(stmt, *args)

    async def fetch_all(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.List[t.Mapping]:
        assert self._connection is not None, "Connection is not acquired."
        stmt, args = self._replace_placeholders(stmt, params)
        return await self._connection.fetch(stmt, *args)

    async def iterate(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.AsyncGenerator[t.Any, None]:
        assert self._connection is not None, "Connection is not acquired."
        stmt, args = self._replace_placeholders(stmt, params)
        async with self.transaction():
            async for row in self._connection.cursor(stmt, *args):
                yield row

    def transaction(self) -> _Transaction:
        return _Transaction(self)

    @property
    def raw_connection(self) -> asyncpg.Connection:
        return self._connection

    async def __aenter__(self) -> "_Connection":
        await self.acquire()
        return self

    async def __aexit__(self, *args) -> None:
        await self.release()

    def _replace_placeholders(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.Tuple[str, t.List]:
        args = []
        if params:
            counter = itertools.count(1)
            for key, value in params.items():
                index = next(counter)
                stmt = stmt.replace(f":{key}", f"${index}")
                args.append(value)
        return stmt, args


class PostgresDriver(
    BaseDriver[pk.dialects.PostgreSQLQuery, pk.dialects.PostgreQueryBuilder]
):
    dialect = "postgresql"
    query_class = pk.dialects.PostgreSQLQuery

    def __init__(self, url: URL) -> None:
        self.url = url
        self.pool: t.Optional[asyncpg.pool.Pool] = None

    async def connect(self) -> None:
        self.pool = asyncpg.create_pool(
            self.url.url,
            **self.url.options,
        )

    async def disconnect(self) -> None:
        assert self.pool, "Driver is not connected."
        await self.pool.close()
        self.pool = None

    def connection(self) -> _Connection:
        return _Connection(self.pool)

    def insert_query(
        self,
        table_name: str,
        values: t.Mapping,
        on_conflict: t.Union[str, t.List[str]] = OnConflict.RAISE,
        conflict_target: t.Union[str, t.List[str]] = None,
        replace_except: t.List[str] = None,
    ) -> pk.dialects.PostgreQueryBuilder:
        qb = super().insert_query(table_name, values)
        conflict_target = conflict_target or []
        replace_except = replace_except or []

        if on_conflict != OnConflict.RAISE:
            qb = qb.on_conflict(*conflict_target)

        if on_conflict == OnConflict.NOTHING:
            qb = qb.do_nothing()

        if on_conflict == OnConflict.REPLACE:
            for column, value in values.items():
                if column in replace_except:
                    continue
                qb = qb.do_update(column, value)
        return qb



import typing as t

import pypika as pk


class OnConflict:
    NOTHING = "nothing"
    REPLACE = "replace"
    RAISE = "raise"


class Raw(pk.terms.Term):
    def __init__(self, raw_sql: str, alias: t.Optional[str] = None) -> None:
        super().__init__(alias=alias)
        self.alias = alias
        self.raw_sql = raw_sql

    def get_sql(self, **kwargs: t.Any) -> str:
        return self.raw_sql



from __future__ import annotations

import asyncio
import typing as t
from types import TracebackType

from aerie.protocols import BaseDriver, Queryable


class Connection:
    def __init__(self, driver: BaseDriver) -> None:
        self._driver = driver
        self._connection = driver.connection()
        self._query_lock = asyncio.Lock()
        self._connection_lock = asyncio.Lock()
        self._connection_counter = 0

        self.transaction_lock = asyncio.Lock()
        self.transaction_counter = 0
        self.transactions: t.List[Transaction] = []

    async def execute(self, stmt: Queryable, params: t.Mapping = None) -> t.Any:
        async with self._query_lock:
            return await self._connection.execute(str(stmt), params)

    async def execute_all(
        self,
        stmt: Queryable,
        params: t.List[t.Mapping] = None,
    ) -> t.Any:
        async with self._query_lock:
            return await self._connection.execute_all(str(stmt), params)

    async def fetch_one(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.Optional[t.Mapping]:
        async with self._query_lock:
            return await self._connection.fetch_one(str(stmt), params)

    async def fetch_all(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.List[t.Mapping]:
        async with self._query_lock:
            return await self._connection.fetch_all(stmt, params)

    async def fetch_val(
        self, stmt: str, params: t.Mapping = None, column: t.Any = 0
    ) -> t.Any:
        async with self._query_lock:
            return await self._connection.fetch_val(stmt, params, column)

    async def iterate(
        self,
        stmt: str,
        params: t.Mapping = None,
    ) -> t.AsyncGenerator[t.Any, None]:
        async with self._query_lock:
            async for row in self._connection.iterate(stmt, params):
                yield row

    def transaction(self) -> Transaction:
        return Transaction(lambda: self)

    async def __aenter__(self) -> Connection:
        async with self._connection_lock:
            self._connection_counter += 1
            if self._connection_counter == 1:
                await self._connection.acquire()
        return self

    async def __aexit__(self, *args) -> None:
        async with self._connection_lock:
            assert self._connection is not None
            self._connection_counter -= 1
            if self._connection_counter == 0:
                await self._connection.release()


class Transaction:
    def __init__(
        self,
        connection_factory: t.Callable[[], Connection],
        force_rollback: bool = False,
    ) -> None:
        self._connection = connection_factory()
        self._transaction = self._connection._connection.transaction()
        self._force_rollback = force_rollback

    @property
    def _is_root(self) -> bool:
        return not self._connection.transaction_counter

    async def begin(self) -> Transaction:
        """Begin transaction."""
        async with self._connection.transaction_lock:
            await self._connection.__aenter__()
            await self._transaction.begin(self._is_root)
            self._connection.transaction_counter += 1
            self._connection.transactions.append(self)
            return self

    async def commit(self) -> None:
        """Commit the transaction."""
        async with self._connection.transaction_lock:
            self._connection.transaction_counter -= 1
            self._connection.transactions.remove(self)
            await self._transaction.commit()
            await self._connection.__aexit__(None, None, None)

    async def rollback(self) -> None:
        """Rollback the transaction."""
        async with self._connection.transaction_lock:
            self._connection.transaction_counter -= 1
            self._connection.transactions.remove(self)
            await self._transaction.rollback()
            await self._connection.__aexit__(None, None, None)

    def __await__(self) -> t.Generator:
        """Creates new transaction instance and automatically begins it.
        You still have to control commit() or rollback() calls.

        Example:
            tx = await db.transaction() # excutes BEGIN under the hood
            tx.commit()
        """
        return self.begin().__await__()

    async def __aenter__(self) -> Transaction:
        return await self.begin()

    async def __aexit__(
        self,
        exc_type: t.Type[BaseException],
        exc_val: BaseException,
        exc_tb: TracebackType,
    ) -> None:
        if self in self._connection.transactions:
            if exc_type is not None or self._force_rollback:
                await self.rollback()
            else:
                await self.commit()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aerie', 'aerie.drivers']

package_data = \
{'': ['*']}

install_requires = \
['PyPika>=0.44.0,<0.45.0']

extras_require = \
{'full': ['aiosqlite>=0.16.0,<0.17.0', 'asyncpg>=0.21.0,<0.22.0'],
 'pgsql': ['asyncpg>=0.21.0,<0.22.0'],
 'sqlite': ['aiosqlite>=0.16.0,<0.17.0']}

setup_kwargs = {
    'name': 'aerie',
    'version': '0.1.0',
    'description': 'A database toolkit.',
    'long_description': None,
    'author': 'alex.oleshkevich',
    'author_email': 'alex.oleshkevich@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import contextvars
import functools
import typing as t

import pypika as pk

from aerie.collections import Collection
from aerie.connection import Connection, Transaction
from aerie.exceptions import DriverNotRegistered
from aerie.protocols import BaseDriver, IterableValues, Queryable
from aerie.terms import OnConflict
from aerie.url import URL
from aerie.utils import import_string

E = t.TypeVar("E", bound=type)

EntityFactoryType = t.Callable[[t.Mapping], t.Any]


def _make_entity(
    map_to: t.Type[E],
    entity_factory: EntityFactoryType,
    row: t.Mapping,
) -> t.Union[t.Mapping, E, t.Any]:
    assert not all(
        [
            map_to is not None,
            entity_factory is not None,
        ]
    ), 'Either "entity_factory" or "map_to" could be passed at same time.'

    if map_to:
        return map_to(**row)

    if entity_factory:
        return entity_factory(row)
    return row


class Database:
    drivers: t.Dict[str, t.Union[str, t.Type[BaseDriver]]] = {
        "sqlite": "aerie.drivers.sqlite.SQLiteDriver",
        "postgres": "aerie.drivers.postgresql.PostgresDriver",
        "postgresql": "aerie.drivers.postgresql.PostgresDriver",
    }

    def __init__(self, url: t.Union[str, URL]) -> None:
        self.url = URL(url) if isinstance(url, str) else url
        self.driver = self.create_driver()
        self._connection_context: contextvars.ContextVar = contextvars.ContextVar(
            "connection"
        )

    async def execute(
        self,
        stmt: Queryable,
        params: t.Mapping = None,
    ) -> t.Any:
        """Execute query with given params."""
        async with self.connection() as connection:
            return await connection.execute(stmt, params)

    async def execute_all(
        self,
        stmt: Queryable,
        params: t.List[t.Mapping] = None,
    ) -> t.Any:
        async with self.connection() as connection:
            return await connection.execute_all(stmt, params)

    async def fetch_one(
        self,
        stmt: Queryable,
        params: t.Mapping = None,
        map_to: t.Type[E] = None,
        entity_factory: EntityFactoryType = None,
        default: t.Union[t.Mapping, E, t.Any] = None,
    ) -> t.Optional[t.Union[t.Mapping, E, t.Any]]:
        factory = functools.partial(_make_entity, map_to, entity_factory)
        async with self.connection() as connection:
            row = await connection.fetch_one(str(stmt), params)
            if row:
                return factory(row)
            return default

    async def fetch_all(
        self,
        stmt: Queryable,
        params: t.Mapping = None,
        map_to: t.Type[E] = None,
        entity_factory: EntityFactoryType = None,
    ) -> t.Union[Collection[t.Mapping], Collection[E], t.Any]:
        factory = functools.partial(_make_entity, map_to, entity_factory)
        async with self.connection() as connection:
            rows = await connection.fetch_all(str(stmt), params)
            return Collection(list(map(factory, rows)))

    async def fetch_val(
        self, stmt: Queryable, params: t.Mapping = None, column: t.Any = 0
    ) -> t.Any:
        async with self.connection() as connection:
            row = await connection.fetch_one(str(stmt), params)
            if row is None:
                return row
            return row[column]

    async def iterate(
        self,
        stmt: Queryable,
        params: t.Optional[t.Mapping] = None,
        map_to: t.Type[E] = None,
        entity_factory: EntityFactoryType = None,
    ) -> t.AsyncGenerator[t.Union[t.Mapping, E, t.Any], None]:
        factory = functools.partial(_make_entity, map_to, entity_factory)
        async with self.connection() as connection:
            async for row in connection.iterate(str(stmt), params):
                yield factory(row)

    async def insert(
        self,
        table_name: str,
        values: t.Mapping,
        on_conflict: str = OnConflict.RAISE,
        conflict_target: t.Union[str, t.List[str]] = None,
        replace_except: t.List[str] = None,
    ) -> int:
        qb = self.driver.insert_query(
            table_name,
            values,
            on_conflict,
            conflict_target,
            replace_except,
        )
        return await self.execute(qb, values)

    async def insert_all(
        self,
        table_name: str,
        values: IterableValues,
    ) -> None:
        qb = self.driver.insert_all_query(table_name, values)
        return await self.execute_all(qb, list(values))

    async def update(
        self,
        table_name: str,
        set: t.Mapping,
        where: t.Union[str, pk.terms.Term] = None,
        where_params: t.Mapping = None,
    ) -> None:
        qb = self.driver.update_query(table_name, set, where)
        where_params = where_params or {}
        await self.execute(qb, {**set, **where_params})

    async def delete(
        self,
        table_name: str,
        where: t.Union[str, pk.terms.Term] = None,
        where_params: t.Mapping = None,
    ) -> None:
        qb = self.driver.delete_query(table_name, where)
        await self.execute(qb, where_params)

    async def upsert(
        self,
        table_name: str,
        values: t.Mapping,
        where: t.Mapping,
    ) -> None:
        ...

    async def count(
        self,
        table_name: str,
        column: str = "*",
        where: t.Union[str, pk.terms.Term] = None,
        where_params: t.Mapping = None,
    ) -> int:
        qb = self.driver.count_query(table_name, column, where)
        return await self.fetch_val(qb, where_params)

    def query_builder(self) -> pk.queries.Query:
        return self.driver.get_query_builder()

    def transaction(self, force_rollback: bool = False) -> Transaction:
        """Manages database transactions (if supported by the used driver).
        It provides context manager interface and
        therefore may be used with `async with` keywords.

        >>> async with db.transaction():
        >>> # do stuff

        Also, for low-level control you can get a transaction instance
        and control its lifecycle by hand.

        >>> tx = db.transaction():
        >>> await tx.begin()
        >>> # do stuff
        >>> await tx.commit()

        There is a shortcut to begin the transaction when creating and object
        to avoid boilerplate:

        >>> tx = await db.transaction():
        >>> # do stuff
        >>> await tx.commit()
        """
        return Transaction(self.connection, force_rollback)

    async def connect(self) -> Database:
        await self.driver.connect()
        return self

    async def disconnect(self) -> None:
        await self.driver.disconnect()

    def create_driver(self) -> BaseDriver:
        if self.url.scheme not in self.drivers:
            raise DriverNotRegistered(
                f'No driver for scheme "{self.url.scheme}". '
                "Use Database.register_driver() to register a new driver."
            )
        driver_class_spec = self.drivers[self.url.scheme]
        if isinstance(driver_class_spec, str):
            driver_class = import_string(driver_class_spec)
        else:
            driver_class = driver_class_spec
        return driver_class(self.url)

    @classmethod
    def register_driver(cls, scheme: str, driver: str) -> None:
        cls.drivers[scheme] = driver

    def connection(self) -> Connection:
        try:
            return self._connection_context.get()
        except LookupError:
            connection = Connection(self.driver)
            self._connection_context.set(connection)
            return connection

    async def __aenter__(self) -> Database:
        return await self.connect()

    async def __aexit__(self, *args) -> None:
        await self.disconnect()

    def __repr__(self) -> str:
        return f"<Database: {self.url}>"
